#define _GNU_SOURCE

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <net/route.h>
#include <stdbool.h>
#include <string.h>
#include <sys/un.h>
#include <signal.h>
#include <net/if.h>
#include <stdio.h>
#include <netdb.h>
#include <sched.h>

#include "sandbox.h"

/*
	Forwards data between two sockets a <-> b and closes both socket handles after the connection is terminated
	(runs as regular user)
*/
static void socketForward(int a, int b){
	fd_set rfds, wfds;
	char bufa[SOCKET_FWD_SIZE]; /* data from b to a */
	char bufb[SOCKET_FWD_SIZE]; /* data from a to b */
	int buflena = 0;
	int buflenb = 0;
	int len;

	/* select maximum file descriptor */
	int nfds = ((a < b) ? b : a) + 1;

	while ( (a >= 0 || b >= 0) &&	/* impossible to deliver data without any connection at all */
			(a >= 0 || buflenb) &&	/* either a still connected or data from a to b */
			(b >= 0 || buflena) ){	/* either b still connected or data from b to a */

		FD_ZERO(&rfds);
		if (a >= 0 && buflenb < SOCKET_FWD_SIZE) FD_SET(a, &rfds);
		if (b >= 0 && buflena < SOCKET_FWD_SIZE) FD_SET(b, &rfds);

		FD_ZERO(&wfds);
		if (a >= 0 && buflena > 0) FD_SET(a, &wfds);
		if (b >= 0 && buflenb > 0) FD_SET(b, &wfds);

		/* wait for socket events */
		if (select(nfds, &rfds, &wfds, NULL, NULL) < 0) break;

		/* read from a (to buffer b) */
		if (a >= 0 && FD_ISSET(a, &rfds) && buflenb < SOCKET_FWD_SIZE){
			len = recv(a, &bufb[buflenb], SOCKET_FWD_SIZE - buflenb, 0);
			if (len <= 0){ close(a); a = (-1); }else{
				buflenb += len;
			}
		}

		/* read from b (to buffer a) */
		if (b >= 0 && FD_ISSET(b, &rfds) && buflena < SOCKET_FWD_SIZE){
			len = recv(b, &bufa[buflena], SOCKET_FWD_SIZE - buflena, 0);
			if (len <= 0){ close(b); b = (-1); }else{
				buflena += len;
			}
		}

		/* write to b */
		if (b >= 0 && FD_ISSET(b, &wfds) && buflenb > 0){
			len = send(b, bufb, buflenb, 0);
			if (len <= 0){ close(b); b = (-1); }else{
				if (len < buflenb) memmove(bufb, &bufb[len], buflenb - len);
				buflenb -= len;
			}
		}

		/* write to a */
		if (a >= 0 && FD_ISSET(a, &wfds) && buflena > 0){
			len = send(a, bufa, buflena, 0);
			if (len <= 0){ close(a); a = (-1); }else{
				if (len < buflena) memmove(bufa, &bufa[len], buflena - len);
				buflena -= len;
			}
		}

	}

	if (a >= 0) close(a);
	if (b >= 0) close(b);
}

/*
	Opens a connection to the path and starts forwarding packets until the connection is terminated by one side.
	(runs as regular user)
*/
static void socketForwardToUnix(int client, const char *path){
	struct sockaddr_un sockAddress;
	int sockfd;

	sockAddress.sun_family = AF_UNIX;

	if (!sec_strncpy(sockAddress.sun_path, path, sizeof(sockAddress.sun_path)))
		goto err;

	sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (sockfd == -1)
		goto err;

	if (connect(sockfd, (struct sockaddr *)&sockAddress, sizeof(struct sockaddr_un)) != 0){
		close(sockfd);
		goto err;
	}

	/* forwards between both sockets (until the connection is closed and the sockets are destroyed) */
	socketForward(client, sockfd);
	return;

err:
	close(client);
}

/*
	Forwards an incoming TCP connection
	(runs as regular user)
*/
static void socketForwardTCP(int client){
	struct sockaddr_storage dest;
	socklen_t destlen = sizeof(dest);
	int sockfd;

	if (getsockname(client, (struct sockaddr *)&dest, &destlen) != 0){
		perror("getsockname() failed");
		goto err;
	}

	sockfd = socket(dest.ss_family, SOCK_STREAM, 0);
	if (sockfd == -1)
		goto err;

	if (connect(sockfd, (struct sockaddr *)&dest, destlen) != 0){
		perror("connect() failed");
		close(sockfd);
		goto err;
	}

	/* forwards between both sockets */
	socketForward(client, sockfd);
	return;

err:
	close(client);
}

/*
	Creates a unix socket for every entry in the linked list
	(RUNS AS ROOT)
*/
bool root_createForwardSocketsUnix(struct childDataStruct *data){
	struct forwardSocketEntry *forwardSocket;

	#ifdef SANDBOXDEBUG
		fprintf(stderr, "[SANDBOX:%d] root_createForwardSocketsUnix(%p)\n", getpid(), data);
	#endif

	for (forwardSocket = data->forwardSocket; forwardSocket; forwardSocket = forwardSocket->next){
		if ((forwardSocket->socket = socket(AF_UNIX, SOCK_STREAM, 0)) == -1){
			perror("Creating socket() failed");
			return false;
		}
	}

	return true;
}

/*
	Creates a TCP forward socket
	(RUNS AS ROOT)
*/
bool root_createForwardSocketTCP(struct childDataStruct *data, int sandboxPid){
	struct sockaddr_in sockAddress_in;
	const int yes = 1;
	bool res = false;
	int oldns;

	#ifdef SANDBOXDEBUG
		fprintf(stderr, "[SANDBOX:%d] root_createForwardSocketTCP(%p)\n", getpid(), data);
	#endif

	/* no TCP forwarding required */
	if (!data->forwardTCP.enabled)
		return true;

	/* switch to the sandbox network namespace, remember the old namespace */
	if ( !root_switchNS(sandboxPid, "net", CLONE_NEWNET, &oldns) ){
		fprintf(stderr, "Unable to join namespaces!\n");
		return false;
	}

	if ((data->forwardTCP.socket = socket(AF_INET, SOCK_STREAM, 0)) == -1){
		perror("Creating socket() failed");
		goto out;
	}

	/* set socket flags */
	if (setsockopt(data->forwardTCP.socket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) != 0){
		perror("Failed to setsockopt(..., SO_REUSEADDR, ...)");
		goto out;
	}

	if (setsockopt(data->forwardTCP.socket, SOL_IP, IP_TRANSPARENT, &yes, sizeof(yes)) != 0){
		perror("Failed to setsockopt(..., IP_TRANSPARENT, ...)");
		goto out;
	}

	memset(&sockAddress_in, 0, sizeof(sockAddress_in));
	sockAddress_in.sin_family  		= AF_INET;
	sockAddress_in.sin_port 		= htons(TPROXY_PORT_TCP);
	sockAddress_in.sin_addr.s_addr	= INADDR_ANY;

	/* bind the TCP socket */
	if (bind(data->forwardTCP.socket, &sockAddress_in, sizeof(sockAddress_in)) != 0){
		perror("Failed to bind() TCP socket");
		goto out;
	}

	/* start listening */
	if (listen(data->forwardTCP.socket, 5) != 0){
		perror("Failed to listen() on TCP socket");
		goto out;
	}

	/* everything okay so far */
	res = true;

out:
	/* switch back to the original network namespace */
	if (!root_setNS(oldns, CLONE_NEWNET))
		res = false;

	return res;
}

/*
	Close all forward sockets
	(RUNS AS ROOT)
*/
void root_closeForwardSockets(struct childDataStruct *data){
	struct forwardSocketEntry *forwardSocket;

	#ifdef SANDBOXDEBUG
		fprintf(stderr, "[SANDBOX:%d] root_closeForwardSockets(%p)\n", getpid(), data);
	#endif

	for (forwardSocket = data->forwardSocket; forwardSocket; forwardSocket = forwardSocket->next){
		if (forwardSocket->socket == (-1))
			continue;

		close(forwardSocket->socket);
		forwardSocket->socket = (-1);
	}

	if (data->forwardTCP.socket != (-1)){
		close(data->forwardTCP.socket);
		data->forwardTCP.socket = (-1);
	}

}

/*
	Binds the forward sockets
	(RUNS AS ROOT!)
*/
bool root_bindForwardSockets(const struct childDataStruct *data){
	struct forwardSocketEntry *forwardSocket;
	struct sockaddr_un sockAddress_un;

	#ifdef SANDBOXDEBUG
		fprintf(stderr, "[SANDBOX:%d] root_bindForwardSockets(%p)\n", getpid(), data);
	#endif

	for (forwardSocket = data->forwardSocket; forwardSocket; forwardSocket = forwardSocket->next){
		sockAddress_un.sun_family = AF_UNIX;

		if (!sec_strncpy(sockAddress_un.sun_path, forwardSocket->path, sizeof(sockAddress_un.sun_path))){
			fprintf(stderr, "Unix socket path '%s' is too long.\n", forwardSocket->path);
			return false;
		}

		/* bind the unix socket */
		if (bind(forwardSocket->socket, (struct sockaddr *)&sockAddress_un, sizeof(struct sockaddr_un)) != 0){
			perror("Failed to bind() unix socket");
			return false;
		}	

		/* adjust file attributes */
		if (chown(forwardSocket->path, forwardSocket->uid, forwardSocket->gid) != 0){
			perror("Failed to chown() unix socket");
			return false;
		}

		if (chmod(forwardSocket->path, forwardSocket->mode & 0777) != 0){
			perror("Failed to chmod() unix socket");
			return false;
		}

		/* start listening */
		if (listen(forwardSocket->socket, 5) != 0){
			perror("Failed to listen() on unix socket");
			return false;
		}

		/* show the final message */
		fprintf(stderr, "[SANDBOX] Forwarding unix socket: %s\n", forwardSocket->path);
	}

	return true;
}

/*
	Starts the unix socket forward daemon
	(RUNS AS ROOT, drops rights after forking)
*/
bool root_forwardSocketDaemon(const struct childDataStruct *data, uid_t realUser, gid_t realGroup, int initPid){
	pid_t daemonPid;

	#ifdef SANDBOXDEBUG
		fprintf(stderr, "[SANDBOX:%d] root_forwardSocketDaemon(%p, %d, %d, %d)\n", getpid(), data, realUser, realGroup, initPid);
	#endif

	/* no daemon necessary if nothing to forward */
	if (!data->forwardSocket && !data->forwardTCP.enabled)
		return true;

	daemonPid = fork();
	if (daemonPid == 0){ /* daemon */

		struct forwardSocketEntry *forwardSocket;
		struct timeval timeout;
		fd_set fds;
		int maxfd, tmp, client;
		pid_t forwardPid;

		/* Drop rights */
		if (!root_dropRights(NULL, realUser, realGroup)){
			fprintf(stderr, "Daemon failed to drop root rights.\n");
			exit(1);
		}

		/* we don't need child notifications, otherwise we will get [defunct] zombies */
		signal(SIGCHLD, SIG_IGN);

		while ( !process_nonchild_terminated(initPid) ){

			FD_ZERO(&fds);
			maxfd = 0;

			/* TCP socket */
			if (data->forwardTCP.enabled){
				FD_SET(data->forwardTCP.socket, &fds);
				if (data->forwardTCP.socket > maxfd) maxfd = data->forwardTCP.socket;
			}

			/* UNIX sockets */
			for (forwardSocket = data->forwardSocket; forwardSocket; forwardSocket = forwardSocket->next){
				FD_SET(forwardSocket->socket, &fds);
				if (forwardSocket->socket > maxfd) maxfd = forwardSocket->socket;
			}

			maxfd++;

			timeout.tv_sec 	= 1;
			timeout.tv_usec = 0;
			
			/* check for events */
			tmp = select(maxfd, &fds, NULL, NULL, &timeout);
			if (tmp < 0)  break;
			if (tmp == 0) continue;

			/* TCP socket */
			if (data->forwardTCP.enabled && FD_ISSET(data->forwardTCP.socket, &fds)){

				/* accept incoming sockets */
				client = accept(data->forwardTCP.socket, NULL, 0);
				if(client < 0)
					continue;

				forwardPid = fork();
				if (forwardPid == 0){
					socketForwardTCP(client);
					exit(0);
				}

				/* close the client here */
				close(client);
			}

			/* UNIX sockets */
			for (forwardSocket = data->forwardSocket; forwardSocket; forwardSocket = forwardSocket->next){
				if (!FD_ISSET(forwardSocket->socket, &fds))
					continue;

				/* accept incoming sockets */
				client = accept(forwardSocket->socket, NULL, 0);
				if(client < 0)
					continue;

				forwardPid = fork();
				if (forwardPid == 0){
					socketForwardToUnix(client, forwardSocket->path);
					exit(0);
				}

				/* close the client here */
				close(client);
			}
		}

		exit(0);

	}else if (daemonPid == -1){ /* error */
		perror("Failed to fork() daemon process");
		return false;
	}

	return true;
}

/*
	Brings a network device up
	(RUNS AS ROOT)
*/
bool root_interfaceUp(const char *name, const struct sockaddr *addr){
	struct ifreq ifr;
	int sock;
	bool res = false;

	#ifdef SANDBOXDEBUG
		fprintf(stderr, "[SANDBOX:%d] root_interfaceUp('%s', %p)\n", getpid(), name, addr);
	#endif

	/* create socket */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock == -1){
		perror("Creating socket() failed");
		return false;
	}

	if (addr){
		/* assign addr */
		memset(&ifr, 0, sizeof(ifr));

		if (!sec_strncpy(ifr.ifr_name, name, IFNAMSIZ))
			goto out;

		memcpy(&ifr.ifr_addr, addr, sizeof(*addr));

		res = (ioctl(sock, SIOCSIFADDR, &ifr) == 0);
		if (!res){
			perror("ioctl() to assign address failed");
			goto out;
		}
	}

	/* bring up device */
	memset(&ifr, 0, sizeof(ifr));

	if (!sec_strncpy(ifr.ifr_name, name, IFNAMSIZ))
		goto out;

	ifr.ifr_flags = IFF_UP;

	res = (ioctl(sock, SIOCSIFFLAGS, &ifr) == 0);
	if (!res){
		perror("ioctl() to bring up device failed");
		goto out;
	}	

out:
	close(sock);
	return res;
}

/*
	Add a default route
	(RUNS AS ROOT)
*/
bool root_setDefaultGateway(const struct sockaddr *addr){
	struct rtentry route;
	struct sockaddr_in *tmp;
	int sock;
	bool res = false;

	#ifdef SANDBOXDEBUG
		fprintf(stderr, "[SANDBOX:%d] root_setDefaultGateway(%p)\n", getpid(), addr);
	#endif

	/* create socket */
	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
	if (sock == -1){
		perror("Creating socket() failed");
		return false;
	}

	/* add route */
	memset(&route, 0, sizeof(route));

	memcpy(&route.rt_gateway, addr, sizeof(*addr));

	tmp = (struct sockaddr_in *)&route.rt_dst;
	tmp->sin_family = AF_INET;
	tmp->sin_addr.s_addr = INADDR_ANY;

	tmp = (struct sockaddr_in *)&route.rt_genmask;
	tmp->sin_family = AF_INET;
	tmp->sin_addr.s_addr = INADDR_ANY;

	route.rt_flags  = RTF_UP | RTF_GATEWAY;
	route.rt_metric = 0;

	res = (ioctl(sock, SIOCADDRT, &route) == 0);
	if (!res){
		perror("ioctl() to set default gateway failed");
		goto out;
	}

out:
	close(sock);
	return res;
}